# == Native Modules
# == Installed Modules
# == Project Modules
import pickle

from build_casoff_input import (check_bulge, parse_bulge)
from scoring import (cfd_score, cfd_spec_score)
from annotate import Transcript


def parse_nobulge(tmp_processing_filename):
	nobulge_dict = {}
	with open(tmp_processing_filename, 'r') as inf:
		for line in inf:
			entry = line.strip().split(' ')
			if len(entry) > 2 and len(entry[-1]) > 3:
				seq, mm, gid = entry
				nobulge_dict[seq] = [gid, mm]
	return nobulge_dict


def write_casoff_output(raw_casoff_out, casoff_support_file, formatted_output, bulge_check):
	with open(raw_casoff_out) as fi, open(formatted_output, 'w') as fo:
		fo.write('Coordinates\tDirection\tGuide_ID\tBulge type\tcrRNA\tDNA\tMismatches\tBulge Size\n') \

		ot_coords = []
		for line in fi:
			entries = line.strip().split('\t')
			ncnt = 0

			if not bulge_check:
				nobulge_dict = parse_nobulge(casoff_support_file)
				gid, mm = nobulge_dict[entries[0]]
				coord = f'{entries[1]}:{entries[2]}-{int(entries[2]) + len(entries[0])}'
				fo.write(f'{coord}\t{entries[4]}\t{gid}\tX\t{entries[0]}\t{entries[3]}\t{entries[5]}\t0\n')
				ot_coords.append(coord)
			if bulge_check:
				(isreversed,
				 chrom_path,
				 seq_pam,
				 rnabulge_dic,
				 id_dict,
				 len_pam,
				 pattern,
				 bulge_dna,
				 bg_tgts) = parse_bulge(casoff_support_file)

				if isreversed:
					for c in entries[0][::-1]:
						if c == 'N':
							ncnt += 1
							break
					if ncnt == 0:
						ncnt = -len(entries[0])
				else:
					for c in entries[0]:
						if c == 'N':
							ncnt += 1
						else:
							break

				if entries[0] in rnabulge_dic:
					gid = id_dict[entries[0]]
					for pos, query_mismatch, seq in rnabulge_dic[entries[0]]:
						if isreversed:
							tgt = (seq_pam + entries[0][len_pam:len_pam + pos] + seq + entries[0][len_pam + pos:-ncnt],
								   entries[3][:len_pam + pos] + '-' * len(seq) + entries[3][len_pam + pos:-ncnt])
						else:
							tgt = (entries[0][ncnt:ncnt + pos] + seq + entries[0][ncnt + pos:-len_pam] + seq_pam,
								   entries[3][ncnt:ncnt + pos] + '-' * len(seq) + entries[3][ncnt + pos:])
						if query_mismatch >= int(entries[5]):
							start = int(entries[2]) + (ncnt if (not isreversed and entries[4] == "+") or (
									isreversed and ncnt > 0 and entries[4] == "-") else 0)
							coord = f'{entries[1]}:{start}-{int(start) + len(tgt[1])}'
							ot_coords.append(coord)
							fo.write(
								f'{coord}\t{entries[4]}\t{gid}\tRNA\t{tgt[0]}\t{tgt[1]}\t{int(entries[5])}\t{len(seq)}\n')

				else:
					gid = id_dict[entries[0]]
					nbulge = 0
					if isreversed:
						for c in entries[0][:-ncnt][len_pam:]:
							if c == 'N':
								nbulge += 1
							elif nbulge != 0:
								break
						tgt = (seq_pam + entries[0][:-ncnt][len_pam:].replace('N', '-'), entries[3][:-ncnt])
					else:
						for c in entries[0][ncnt:][:-len_pam]:
							if c == 'N':
								nbulge += 1
							elif nbulge != 0:
								break
						tgt = (entries[0][ncnt:][:-len_pam].replace('N', '-') + seq_pam, entries[3][ncnt:])
					start = int(entries[2]) + (ncnt if (not isreversed and entries[4] == "+") or (
							isreversed and ncnt > 0 and entries[4] == "-") else 0)
					btype = 'X' if nbulge == 0 else 'DNA'
					coord = f'{entries[1]}:{start}-{int(start) + len(tgt[1])}'
					ot_coords.append(coord)
					fo.write(
						f'{entries[1]}:{start}-{start + len(tgt[1])}\t{entries[4]}\t{gid}\t{btype}\t{tgt[0]}\t{tgt[1]}\t{int(entries[5])}\t{nbulge}\n')

		editor = gid.split('_')[0]
		print(f'{len(ot_coords)} off targets found for {editor}')


def score_ot(crrna,otseq,models_dir):
	score = 0
	if '-' not in crrna[:-3]:
		if '-' not in otseq[:-4]:
			# TODO: NEEDS
			#raise Exception("The function scoring.cfd_score requires a path to the model files")
			score = cfd_score(crrna[:-3].upper(), otseq.upper(),models_dir)
	return score


def annotate_ots(output_filename, annote_path, models_dir):
	'''
	Reads output, Scores each off-target seq and annotates each off_target
	'''

	editor = output_filename.split('_casoff')[0].split('_')[-1]
	coords = []
	scores = []
	spec_scores = {}
	res = open(output_filename, 'r').readlines()
	for line in res[1:]:
		line = line.strip().split('\t')
		coords.append(line[0])
		if line[2] not in spec_scores.keys():
			spec_scores[line[2]] = 0 if editor == 'spCas9' else '.'
		if editor == 'spCas9':
			score = score_ot(line[4], line[5],models_dir)
			if score > 0 and score != 1:
				spec_scores[line[2]] = spec_scores[line[2]] + score
		else:
			score = '.'
		scores.append(score)

	Transcript.load_transcripts(annote_path,coords)

	new_lines = []
	annotate_out = output_filename.replace('_output', '_annotated')
	with open(annotate_out, 'w') as anout:
		anout.write(res[0].strip() + f'\tAnnotation\tScore\n')
		cnt = 0
		for line in res[1:]:
			line = line.strip().split('\t')
			ans = Transcript.transcript(line[0])

			if ans != 'intergenic':
				tid, gname = ans.tx_info()[1:3]
				feature = ans.feature
				x = '|'.join([feature, gname, tid])
				new_line = line + [x, str(scores[cnt])]
				new_lines.append(new_line)

			else:
				x = 'intergenic'
				new_line = line + [x, str(scores[cnt])]
				new_lines.append(new_line)
			cnt += 1
			anout.write('\t'.join(new_line) + '\n')

	if editor == 'spCas9':
		for gid, sum_score in spec_scores.items():
			if sum_score!= 0:
				spec_scores[gid] = cfd_spec_score(sum_score)
		else:
			spec_scores[gid] = "."
	#remove(output_filename)
	return new_lines, spec_scores


def agg_results(lines,mmco):
	'''
	sums the number of off-targets for each guide
	aggregates by mismatch cutoff and bulge size
	'''
	ots_dict = {}
	for line in lines:
		gid, btype, mm, bsize = line[2], line[3], line[6], line[7]
		if gid not in ots_dict.keys():
			ots_dict[gid] = {}
			for i in ['X', 'RNA', 'DNA']:
				for j in range(mmco+1):
					ots_dict[gid][(i,j)] = 0
		ots_dict[gid][(btype, int(mm))] += 1
	return ots_dict


def main():
	# SNAKEMAKE IMPORTS
	# === Inputs ===
	raw_casoff_output_path = str(snakemake.input.casoff_out)
	casoff_support_path = str(snakemake.input.casoff_support)
	# === Outputs ===
	offtarget_scores = str(snakemake.output.offtarget_scores)
	formatted_casoff_temp = str(snakemake.output.formatted_casoff_temp)
	# === Params ===
	rna_bulge = str(snakemake.params.rna_bulge)
	dna_bulge = str(snakemake.params.dna_bulge)
	maximum_mismatches = str(snakemake.params.max_mismatch)
	PU = str(snakemake.params.casoff_accelerator)
	annote_path = str(snakemake.params.annote_path)
	models_path = str(snakemake.params.models_path)

	# Check bulge based on pre-defined Cas-Offinder params
	casoff_params = (maximum_mismatches, rna_bulge, dna_bulge, PU)
	bulge_check = check_bulge(casoff_params)

	write_casoff_output(raw_casoff_output_path,
						casoff_support_path,
						formatted_casoff_temp,
						bulge_check)

	# == Process off-target scoring
	new_lines, spec_scores = annotate_ots(formatted_casoff_temp, annote_path, models_path)
	ot_dict = agg_results(new_lines, casoff_params[0])  # sum off-targets

	# == Compile and export scores
	ots = {}
	for k, v in ot_dict.items():
		v['spec_score'] = spec_scores[k]
		ots[k] = v
	with open(offtarget_scores, 'wb') as pickle_handle:
		pickle.dump(ots, pickle_handle)


if __name__ == "__main__":
	main()



# == Native Modules
import re
import sys
import sys
# == Installed Modules
from Bio import SeqIO
# == Project Modules


def main():
	INPUT_FASTA = sys.argv[1]

	record = SeqIO.parse(open(INPUT_FASTA), "fasta")
	changed_recs = []
	for rec in record:
		a = re.search(r"Homo sapiens chromosome (\d+), \S+ Primary Assembly", rec.description)
		try:
			rec.id = f"chr{a.group(1)}"
			rec.name = f"chr{a.group(1)}"
			rec.description = ""
		except AttributeError:
			pass
		changed_recs.append(rec)

		SeqIO.write(changed_recs, INPUT_FASTA, 'fasta')


if __name__ == "__main__":
	main()



# == Native Modules ==
import subprocess
import gzip
import shutil
from datetime import datetime
import secrets
import string
import pytz
import os
import os.path
import re
import requests
import pathlib
from pathlib import Path
import pickle
# == Installed Modules ==
from alive_progress import alive_bar
import yaml
from importlib_resources import files
from Bio import SeqIO
from Bio.Data import IUPACData
import boto3
from botocore.exceptions import NoCredentialsError
from botocore import UNSIGNED
from botocore.config import Config

import pandas as pd
# == Project Modules ==


def compress_file(file_path: str):
	if not is_gzipped(file_path):
		# If not gzipped, compress the file
		with open(file_path, 'rb') as f_in, gzip.open(file_path + '.gz', 'wb') as f_out:
			shutil.copyfileobj(f_in, f_out)
		print(f"File '{file_path}' compressed successfully.")
	if is_gzipped(file_path):
		cmd_rename = f"mv {file_path} {file_path}.gz"
		subprocess.run(cmd_rename, shell=True)
		print("This file is already compressed.")
		print(f"Created a copy of the file input on: {file_path}.gz")


def consolidate_s3_download(content, parent_folder):
	consolidated_downloadable = []
	for content_idx in range(0, len(content)):
		key = content[content_idx]['Key']
		if key != f"{parent_folder}/":
			consolidated_downloadable.append(content[content_idx])
	return consolidated_downloadable


def check_format(variable, data_type, paramater_name, default_value):
	if isinstance(variable, data_type):
		if variable == default_value:
			print(f"Please specify a value for the option --{paramater_name}.")
			exit(0)
		if data_type == str:
			if not check_iupac(variable):
				print(f"The string provided in --{paramater_name} is not a valid IUPAC representation: {variable}")
				exit(0)
		return data_type(variable)
	else:
		if data_type != str:
			try:
				data_type(variable)
				return data_type(variable)
			except ValueError:
				print(f"Invalid data type for {paramater_name}: '{variable}'. Please double-check the documentation.")
				exit(0)


def check_iupac(sequence):
	# Get the valid IUPAC characters (both unambiguous and ambiguous nucleotides)
	iupac_nucleotides = set(IUPACData.ambiguous_dna_letters)

	# Convert the sequence to uppercase and check if all characters are valid
	return all(base in iupac_nucleotides for base in sequence.upper())


def date_tag():
	# Create a random string of 20 characters
	random_str = ''.join(secrets.choice(string.ascii_letters + string.digits) for _ in range(10))

	# Set the timezone to PST
	pst = pytz.timezone('America/Los_Angeles')
	# Get the current date and time
	current_datetime = datetime.now(pst)
	# Format the date as a string with day, hour, minute, and second
	formatted_date = f"{current_datetime.strftime('%y%m%d%H%M%S%f')}_{random_str}"

	return formatted_date


def download_s3_objects(s3_bucket_name: str, s3_object_name: str, destination_path: str):
	"""
	Downloads a file or every file in a given AWS S3 bucket
	:param s3_bucket_name:
	:param s3_object_name:
	:param destination_path:
	:return:
	"""
	s3 = boto3.client('s3', config=Config(signature_version=UNSIGNED))
	try:
		response = s3.list_objects_v2(Bucket=s3_bucket_name, Prefix=s3_object_name)
		content = response.get('Contents', [])
		clean_content = consolidate_s3_download(content, s3_object_name)  # Skip the S3 object that denotes the parent folder
	except NoCredentialsError:
		prRed("AWS issued credentials error. This is not an expected behavior. Please notify log this error on GitHub")
		exit(0)
	with alive_bar(len(clean_content), title=f'Downloading s3://{s3_bucket_name}/{s3_object_name}') as bar:
		for content_idx in range(0, len(clean_content)):
			key = clean_content[content_idx]['Key']
			source_file = key.split("/")[-1]
			destination_file = os.path.join(destination_path, source_file)
			if file_exists(destination_file):
				print(f"Skipping existing file: {destination_file}")
				bar()
				continue
			# == This downloads the data from the S3 Bucket without requiring AWS credentials
			pathlib.Path(destination_path).mkdir(parents=True, exist_ok=True)
			s3.download_file(s3_bucket_name, key, destination_file)
			bar()


def export_guides_by_editor(guide_df_by_editor_dict: dict, output_dir: (str, Path)):
	editors_list = []
	for editor in guide_df_by_editor_dict:
		editors_list.append(editor)
		guide_df = pd.DataFrame(guide_df_by_editor_dict[editor][0])
		filepath = f"{output_dir}{editor}.pkl"
		# Create output directory if non-existent
		set_export(output_dir)
		with open(filepath, 'wb') as guide_df_handle:
			pickle.dump(guide_df, guide_df_handle)
	return editors_list


def file_exists(file_path):
	return os.path.exists(file_path)


def group_guide_table(guide_table_path: pathlib.Path, editor_filter: (list, str)):
	guides_df = pd.read_csv(guide_table_path)
	if editor_filter:
		guides_df = guides_df[guides_df['Editor'].isin(editor_filter)]
	try:
		grouped_guides_df = guides_df.groupby('Editor')
	except KeyError:
		print("Column name 'Editor' not found in <Guides_found.csv>. Please check the file path and try again.")
		exit(0)
	editor_expanded_dictionary = {}
	for editor, stats in grouped_guides_df:
		editor_expanded_dictionary.setdefault(editor, []).append(stats)
	return editor_expanded_dictionary


def handle_shell_exception(subprocess_result, shell_command, verbose: bool):
	# === Silverplate Errors
	#   == File not found
	if re.search("FileNotFound", str(subprocess_result)):
		prRed(f"QUERY FILEPATH NOT FOUND: Please note:"
			  f"- This is the invalid filepath: {shell_command}"
			  f"- If providing more than one query path make sure the paths are comma-separated")
		exit(0)
	#	== Config Not Found
	if re.search("ConfigNotFound", str(subprocess_result)):
		prRed(f" Couldn't find the file {shell_command}."
			  f" Please double-check the path to <medit_database> and provide the path via '-d' option")
		exit(0)
	# === Handle SMK exceptions through subprocess
	#   == Unlock directory if necessary for SMK run
	if re.findall("Directory cannot be locked.", subprocess_result.stdout) or re.findall("Directory cannot be locked.", subprocess_result.stderr):
		print("--> Target directory locked. Unlocking...")
		unlock_smk_command = f"{shell_command} --unlock"
		launch_shell_cmd(unlock_smk_command, verbose)
		launch_shell_cmd(shell_command, verbose)
		return
	#	== Handle Missing Output Error in Snakemake
	if re.findall("MissingOutputException", subprocess_result.stdout):
		prRed("--> The association between Inputs/Outputs is not valid. Outputs may not be generated")
		return
	#   == Skipping rule call that has already been completed
	if re.findall(r"ValueError: min\(\) arg is an empty sequence", subprocess_result.stderr):
		print("--> A consensus FASTA has already been generated for this job. Skipping.")
		return
	if not re.findall(r"ValueError: min\(\) arg is an empty sequence", subprocess_result.stderr):
		if verbose:
			print(subprocess_result.stderr)
			prGreen(subprocess_result.stdout)
		return


def is_bgzipped(file_path: str) -> bool:
	with open(file_path, 'rb') as f:
		# Check if the file starts with the bgzip magic bytes
		# BGZF magic bytes: 1f 8b 08 04
		return f.read(4) == b'\x1f\x8b\x08\x04'


def is_gzipped(file_path: str):
	with open(file_path, 'rb') as f:
		# Check if the file starts with the gzip magic bytes
		return f.read(2) == b'\x1f\x8b'


def launch_shell_cmd(command: str, verbose=False, **kwargs):
	message = kwargs.get('message', False)
	check_exist = kwargs.get('check_exist', False)
	if message:
		verbose = False
		print(message)
	if check_exist:
		if os.path.isfile(check_exist):
			print(f"File {check_exist} exists. Skipping process.")
			return
	if verbose:
		prCyan(f"--> Invoking command-line call:\n{command}")

	result = subprocess.run(command,
	                        shell=True,
	                        stderr=subprocess.PIPE,
	                        stdout=subprocess.PIPE,
							text=True,
	                        universal_newlines=True
	                        )
	handle_shell_exception(result, command, verbose)


def list_files_by_extension(root_path, extension: str):
	file_list = []
	for root, dirs, files in os.walk(root_path, topdown=False):
		for name in files:
			if name.endswith(extension):
				file_list.append(os.path.join(root, name))
	return file_list


def offtarget_mode_formatting(mode, dynamic_config_guidepred):
	"""
	This function retrieves the relevant sequence IDs of VCFs and reference
	genome to help build the offtarget configuration file
	:param mode: current mEdit run mode
	:param dynamic_config_guidepred: YAML-imported Config dict
	:return: Tuple of size n>=1 with the sequence IDs and respective labels: Either 'main_ref' or 'extended'
	"""
	sequence_id = [(dynamic_config_guidepred['sequence_id'][0]), 'main_ref']
	if mode != 'fast':
		sequence_id.extend((dynamic_config_guidepred['vcf_id'], 'extended_ref'))
	return sequence_id


def pickle_chromosomes(genome_fasta, output_dir):
	records = SeqIO.parse(open(genome_fasta, 'rt'), "fasta")
	with alive_bar(25, title=f'Serializing human chromosomes') as bar:
		for record in records:
			if re.search(r"chr\w{0,2}$", record.id, re.IGNORECASE):
				outfile = f"{output_dir}/{record.id}.pkl"
				with open(outfile, 'ab') as gfile:
					pickle.dump(record, gfile)
					bar()


def prCyan(skk):
	print("\033[96m {}\033[00m".format(skk))


def prGreen(skk):
	print("\033[92m {}\033[00m".format(skk))


def prRed(skk):
	print("\033[0;31;47m {}\033[00m".format(skk))


def project_file_path(path_from_toplevel: str, filename: str):
	"""
	There are two top-level directories in the current version of mEdit: snakemake and config
	From either of these paths, the respective *.smk and *.yaml files can be accessed
	:param path_from_toplevel:
	:param filename:
	:return:
	"""
	return str(files(path_from_toplevel).joinpath(filename))


def set_export(outdir: str):
	if os.path.exists(outdir):
		pass
		# print(f'--> Skipping directory creation: {outdir}')
	# Create outdir only if it doesn't exist
	if not os.path.exists(outdir):
		print(f'Directory created on: {outdir}')
		pathlib.Path(outdir).mkdir(parents=True, exist_ok=True)
	return outdir


def validate_editor_list(editor_request: str, built_in_editors: list, parameter_string: str):
	validated_editors = []
	editor_list = editor_request.split(",")
	for editor in editor_list:
		green_light = False
		for built_in_editor in built_in_editors:
			clean_editor= editor.strip()
			if re.search(clean_editor, built_in_editor, re.IGNORECASE):
				green_light = True
				validated_editors.append(built_in_editor)
		if not green_light:
			print(
				f"Full Set of Refs: {built_in_editors}"
				f"Please call 'medit list --help' to see directions on how to obtain the current list of editors.\n"
				f"Alternatively, use the '{parameter_string} custom' parameter to customize your own editor for this run.\n")
			exit(0)
	return validated_editors


def write_yaml_to_file(py_obj, filename: str):
	with open(f'{filename}', 'w', ) as f:
		yaml.safe_dump(py_obj, f, sort_keys=False, default_style='"')
	print(f'--> Configuration file created: {filename}')



# == Native Modules ==
from argparse import ArgumentParser as argp
from argparse import RawTextHelpFormatter
import textwrap


# == Installed Modules ==
from importlib.metadata import version


def parse_arguments():
	# -> === Launch argparse parser === <-
	parser = argp(
		prog='mEdit',
		description=f'version {version("meditability")}',
		# epilog="mEdit is pretty cool, huh? :)",
		usage='%(prog)s ',
		formatter_class=RawTextHelpFormatter
	)

	programs = parser.add_subparsers(
		title="== mEdit Programs ==",
		description=textwrap.dedent('''
		mEdit can be operated through a list of different programs.\n'''),
		dest="program",
	)
	# === Db Setup ===
	dbset_parser = programs.add_parser(
		'db_set',
		help=textwrap.dedent('''
			Setup the necessary background data to run mEdit\n'''),
		formatter_class=RawTextHelpFormatter
	)
	ref_db_parse = dbset_parser.add_argument_group("== Reference Database Pre-Processing ==")
	ref_db_parse.add_argument('-d',
							  dest='db_path',
							  default='.',
							  help=textwrap.dedent('''
	                          Provide the path where the "mEdit_database" 
	                          directory will be created ahead of the analysis.
	                          Requires ~3.2GB in-disk storage 
	                          [default: ./mEdit_database]\n'''))
	ref_db_parse.add_argument('-l',
							  dest='latest_reference',
							  action='store_true',
							  help=textwrap.dedent('''
							  Request the latest human genome reference as part
							  of mEdit database unpacking. This is especially 
							  recommended when running predictions on private 
							  genome assemblies. [default: False]\n'''))
	ref_db_parse.add_argument('-c',
							  dest='custom_reference',
							  help=textwrap.dedent('''
							  Provide the path to a custom human reference 
							  genome  in FASTA format. 
							  ***Chromosome annotation must follow a
							  ">chrN" format (case sensitive)\n'''))
	ref_db_parse.add_argument('-t',
							  dest='threads',
							  default='1',
							  help=textwrap.dedent('''
	                          Provide the number of cores for parallel 
	                          decompression of mEdit databases.
	                          \n'''))

	# === Editors List ===
	list_parser = programs.add_parser(
		'list',
		help=textwrap.dedent('''
				Prints the current set of editors available on mEdit\n'''),
		formatter_class=RawTextHelpFormatter
	)
	editors_list = list_parser.add_argument_group("== Available Editors and BEs ==")
	editors_list.add_argument('-d',
							  dest='db_path',
							  default='.',
							  help=textwrap.dedent('''
	                          Provide the path where the "mEdit_database"
	                          directory was created ahead of the analysis
	                          using the "db_set" program.
	                          [default: ./mEdit_database]\n''')
							  )
	# editors_list.add_argument(('--editors'),
	# 						  dest='editors',
	# 						  action='store_true',
	# 						  help=textwrap.dedent('''
	#                           Provides the current list of available editors on mEdit
	#                            \n'''))
	# editors_list.add_argument('-b',
	# 						  dest='base_editors',
	# 						  action='store_true',
	# 						  help=textwrap.dedent('''
	#                           Provides the current list of available base editors on mEdit
	#                           \n'''))

	# === Guide Prediction Program ===

	fguides_parser = programs.add_parser(
		'guide_prediction',
		help=textwrap.dedent('''
			The core mEdit program finds potential guides for
			variants specified on the input by searching a diverse set of
			editors.\n'''),
		formatter_class=RawTextHelpFormatter
	)
	in_out = fguides_parser.add_argument_group("== Input/Output Options ==")
	in_out.add_argument(
		'-i',
		dest='query_input',
		required=True,
		help=textwrap.dedent('''
			Path to plain text file containing the query (or set of queries) 
			of variant(s) for mEdit analysis. 
			See --qtype for formatting options.
			\n''')
	)
	in_out.add_argument(
		'-o',
		dest='output',
		default='medit_analysis',
		help=textwrap.dedent('''
			Path to root directory where mEdit outputs will be stored 
			[default: mEdit_analysis_<jobtag>/]\n''')
	)
	in_out.add_argument('-d',
						dest='db_path',
						default='.',
						help=textwrap.dedent('''
	                    Provide the path where the "mEdit_database" 
	                    directory was created ahead of the analysis 
	                    using the "db_set" program. 
	                    [default: ./mEdit_database]\n''')
						)
	in_out.add_argument('-j',
						dest='jobtag',
						help=textwrap.dedent('''
	                    Provide the tag associated with the current mEdit job.
	                    mEdit will generate a random jobtag by default\n''')
						)
	run_params = fguides_parser.add_argument_group("== mEdit Core Parameters ==")
	run_params.add_argument(
		'-m',
		dest='mode',
		default='standard',
		choices=['fast', 'standard', 'vcf'],
		help=textwrap.dedent('''
			The MODE option determines how mEdit will run your job. 
			[default = "standard"]
			[1-] "fast": will find and process guides based only on one 
			reference human genome.
			[2-] "standard": will find and process guides based on a 
			reference human genome assembly along with a diverse set of 
			pangenomes from HPRC.
			[3-] "vcf": requires a custom VCF file that will be 
			processed for guide prediction.\n''')
	)
	run_params.add_argument(
		'-v',
		dest='custom_vcf',
		default=None,
		help=textwrap.dedent('''
			Provide a gunzip compressed VCF file to run mEdit’s 
			vcf mode\n''')
	)
	run_params.add_argument(
		'--qtype',
		dest='qtype_request',
		default='hgvs',
		choices=['hgvs', 'coord'],
		help=textwrap.dedent('''
			Set the query type provided to mEdit. [default = "hgvs"]
			[1-] "hgvs": must at least contain the Refseq identifier 
			followed by “:” and the commonly used HGVS nomenclature. 
			Example: NM_000518.5:c.114G>A, 
			NM_000518.5(HBB):c.114G>A(p.Trp38Ter), 
			NG_000007.3:g.70838G>A, 
			NC_000011.10:g.5226778C>T
			[2-] "coord": must contain hg38 1-based coordinates followed by 
			(ALT>REF). Alleles must be the plus strand.
			Example: chr11:5226778C>T
			\n\n''')
	)
	run_params.add_argument(
		'--editor',
		dest='editor_request',
		default='clinical',
		help=textwrap.dedent('''
			Delimits the set of editors to be used by mEdit. 
			[default = "clinical"]
			[1-] "clinical": this value calls for a short list of clinically
			relevant editors that are either in pre-clinical or 
			clinical trials.
			[2-]<user defined list>: The user provides a comma-separated 
			list of editors. Use the 'medit list' command to access the current 
			set of available editors.
			[3-] "custom": Apply custom guide search parameters. 
			This requires a separate input of parameters: 
			‘--pam’, ‘--pamisfirst’,’--guidelen’, and 'dsb_pos'
			\n''')
	)
	run_params.add_argument(
		'--be',
		dest='be_request',
		default='default',
		help=textwrap.dedent('''
		Add this flag to allow mEdit process base-editors. [default = off]
		[1-] “off”: disable base editor guides searching.
		[2-] “default”: use generic ABE and CBE with ‘NGG’ PAM 
		and 4-8 base editing window
		[3-] “custom”: : select base editor search parameters. 
		This requires a separate input of parameters :
		‘--pam’, ‘--pamisfirst’,’--guidelen’,’--edit_win’,
		’--target_base’, and ’--result_base’
		[4-]<user defined list>: The user provides a comma-separated 
		list of base editors. Use 'medit list' to access the current set of
		available editors.
		\n''')
	)
	run_params.add_argument(
		'--cutdist',
		dest='cutdist',
		default='7',
		help=textwrap.dedent('''
					Max allowable window a variant start position can be from
					the editor cut site. This option not available for 
					base editors. 
					[default = 7]\n''')
	)
	run_params.add_argument(
		'--dry',
		dest='dry_run',
		action='store_true',
		help=textwrap.dedent('''
				Perform a dry run of mEdit.\n'''))

	custom_options = fguides_parser.add_argument_group("== Custom Editor Options ==")
	custom_options.add_argument(
		'--pam',
		dest='pam',
		default='XXX',
		help=textwrap.dedent('''
	        Specifies the PAM sequence to be used for custom guide or 
	        base editor searches. Required if "--editor custom" or
	        "--be custom" is used.
	    \n''')
	)
	custom_options.add_argument(
		'--guidelen',
		dest='guide_length',
		default=-1,
		help=textwrap.dedent('''
	        Specifies the length of the guide sequence for both 
	        custom endonuclease and base editor searches.
	        Required if "--editor custom" or "--be custom" is used.
	    \n''')
	)
	custom_options.add_argument(
		'--pamisfirst',
		dest='pam_is_first',
		action='store_true',
		help=textwrap.dedent('''
	        Indicates if the PAM is positioned before the guide for both
	        custom endonuclease and base editor searches. 
	        Required if "--editor custom" or "--be custom" is used.
	    \n''')
	)
	custom_options.add_argument(
		'--dsb_pos',
		dest='dsb_position',
		default=-10000,
		help=textwrap.dedent('''
	        Double strand cut site relative to pam. This can be a single
	        integer with a blunt end endonuclease or 2 integers separated 
	        by a single comma when using an endonuclease that produces 
	        staggered end cuts. For example spCas9 would be “-3” and 
	        Cas12 is “18,22”.
	        Required if "--editor custom" is used.
	    \n''')
	)
	custom_options.add_argument(
		'--edit_win',
		dest='editing_window',
		default=(0, 0),
		help=textwrap.dedent('''
	        Specifies the size of the editing window for custom base editor
	        search. Two positive integers separated by a comma that represent
	        the base editing window. The numbering begins at the 5’ most end.
	        For example: CBE window is “4,8"
	        Required if "--be custom" is used.
	    \n''')
	)
	custom_options.add_argument(
		'--target_base',
		dest='target_base',
		choices=['A', 'C', 'G', 'T'],
		default='X',
		help=textwrap.dedent('''
	        Specifies the target base for base editor modification.
	        For example: ABE target base is “A”
	        Required if "--be custom" is used.
	    \n''')
	)
	custom_options.add_argument(
		'--result_base',
		dest='result_base',
		choices=['A', 'C', 'G', 'T'],
		default='X',
		help=textwrap.dedent('''
	        Specifies the base that the target base will be converted
	        to for base editor searches. For example: ABE result base is “G”
	        Required if "--be custom" is used.
	    \n''')
	)

	cluster_opt = fguides_parser.add_argument_group("== SLURM Options ==")
	cluster_opt.add_argument(
		'-p',
		dest='parallel_processes',
		help=textwrap.dedent('''
				Most processes in mEdit can be submitted to SLURM.
				When submitting mEdit jobs to SLURM, the user can specify
				the number of parallel processes that will be sent to the 
				server [default = 1]\n''')
	)
	cluster_opt.add_argument(
		'--ncores',
		dest='ncores',
		default=2,
		help=textwrap.dedent('''
			Specify the number of cores through which each parallel process 
			will be computed. [default = 2]\n''')
	)

	cluster_opt.add_argument(
		'--maxtime',
		dest='maxtime',
		default='1:00:00',
		help=textwrap.dedent('''
			Specify the maximum amount of time allowed for each parallel job.
			Format example: 2 hours -> "2:00:00" [default = 1 hour]\n''')
	)

	# === Off Target Effect Program ===
	casoff_parser = programs.add_parser(
		'offtarget',
		help=textwrap.dedent('''
			Predict off-target effect for the guides found\n'''),
		formatter_class=RawTextHelpFormatter
	)
	offtarget_params = casoff_parser.add_argument_group("== Off-Target Parameters ==")
	offtarget_params.add_argument(
		'--dry',
		dest='dry_run',
		action='store_true',
		help=textwrap.dedent('''
				Perform a dry run of mEdit.\n'''))

	off_in_out = casoff_parser.add_argument_group("== Input/Output Options ==")
	off_in_out.add_argument(
		'-o',
		dest='output',
		default='medit_analysis',
		help=textwrap.dedent('''
		Path to root directory where mEdit guide_prediction
		outputs were stored. "medit offtarget" can't 
		operate if this path is incorrect. [default: mEdit_analysis_<jobtag>/]
		 \n''')
	)
	off_in_out.add_argument('-d',
							dest='db_path',
							default='.',
							help=textwrap.dedent('''
		                    Provide the path where the "mEdit_database" 
		                    directory was created ahead of the analysis 
		                    using the "db_set" program. 
		                    [default: ./mEdit_database]\n''')
							)
	off_in_out.add_argument('-j',
							dest='jobtag',
							required=True,
							help=textwrap.dedent('''
							Provide the tag associated with the desired 
							"medit guide_prediction" job ID.
							"mEdit offtarget" will use the path from the
							 OUTPUT option to access this JOBTAG.
							\n''')
							)
	off_in_out.add_argument('--select_editors',
							dest='select_editors',
							default='',
							help=textwrap.dedent('''
								Provide a comma-separated list to select which 
								editors should be analyzed for offtarget effect.
								[default: all] 
								\n''')
							)

	off_cluster_opt = casoff_parser.add_argument_group("== SLURM Options ==")
	off_cluster_opt.add_argument(
		'-p',
		dest='parallel_processes',
		help=textwrap.dedent('''
					Most processes in mEdit can be submitted to SLURM.
					When submitting mEdit jobs to SLURM, the user can specify
					the number of parallel processes that will be sent to the 
					server [default = 1]\n''')
	)
	off_cluster_opt.add_argument(
		'--ncores',
		dest='ncores',
		default=2,
		help=textwrap.dedent('''
				Specify the number of cores through which each parallel 
				process will be computed. [default = 2]\n''')
	)

	off_cluster_opt.add_argument(
		'--maxtime',
		dest='maxtime',
		default='1:00:00',
		help=textwrap.dedent('''
				Specify the maximum amount of time allowed for each parallel job.
				Format example: 2 hours -> "2:00:00" [default = 1 hour]\n''')
	)
	# TODO: Finish the other options at the user interface

	# Parse arguments from the command line
	arguments = parser.parse_args()
	return arguments



# == Native Modules
from os.path import abspath
import pickle
import subprocess
from pathlib import Path
# == Installed Modules
import yaml
# == Project Modules
from prog.medit_lib import (
	export_guides_by_editor,
	file_exists,
	group_guide_table,
	launch_shell_cmd,
	offtarget_mode_formatting,
	project_file_path,
	set_export,
	write_yaml_to_file
)


def offtarget_prediction(args, jobtag):
	# == Load Run Parameters values ==
	user_jobtag = args.user_jobtag
	root_dir = abspath(args.output)
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	editing_tool_request = args.select_editors
	# == Load SLURM-related values ==
	ncores = args.ncores
	maxtime = args.maxtime
	parallel_processes = args.parallel_processes
	dry_run = args.dry_run

	# == Set export paths tied to the SMK pipeline ==
	config_dir_path = f"{root_dir}/config"

	if editing_tool_request:
		editing_tool_request = list(editing_tool_request.split(","))

	# == Set import/export paths for dynamic YAML files ==
	config_db_path = f"{db_path_full}/config_db/config_db.yaml"
	dynamic_config_off_path = f"{config_dir_path}/config_off_{jobtag}.yaml"  # OffTarget config generated in this program
	dynamic_config_guidepred_path = f"{config_dir_path}/config_{jobtag}.yaml"  # Main config generated by guide_prediction.py

	# == Check existence of the config file from guide_prediciton ==
	if not file_exists(dynamic_config_guidepred_path):
		raise (f"This program depends on the configuration file associated with the JOBTAG {jobtag}. "
			   f"Such file wasn't found on this parent folder: {root_dir}. "
			   f"Make sure to provide the same path as the OUTPUT argument given to 'medit guide_prediction' ")

	#   == Load template configuration files ==
	config_template_path = project_file_path("smk.config", "medit_offtarget.yaml")
	config_cluster_path = project_file_path("smk.config", "medit_cluster.yaml")

	# == Define dynamic SMK call variables ==
	cluster_smk_setup = ''
	smk_verbosity = [True]
	smk_run_triggers = ''
	dryrun_setup = ''

	#   == Check the dry run request
	if dry_run:
		dryrun_setup = '-n'
	if user_jobtag:
		smk_run_triggers = '--rerun-triggers "mtime"'
	#   == Check the request to run on a cluster
	if parallel_processes:
		# --> Upon SLURM run request, the guide_prediction.smk is split in two separate runs
		# --> That's because samtool's conda package crashes on a libcrypto error when
		#       it's deployed by snakemake on a SLURM node
		cluster_smk_setup = ('--cluster "sbatch -t {cluster.time} -n {cluster.cores}" '
							 f'--cluster-config  {config_cluster_path}')
	# == Define dynamic SMK call variable ==
	allowed_rules = ['']

	# ->=== CONFIG FILES IMPORT ===<-
	with open(dynamic_config_guidepred_path, 'r') as dynamic_config_handle:
		dynamic_config_guidepred = yaml.safe_load(dynamic_config_handle)
	with open(config_template_path, 'r') as config_handle:
		config_template = yaml.safe_load(config_handle)

	# === Import Variables from Configuration File ===
	run_name = str(dynamic_config_guidepred['run_name'])
	mode = str(dynamic_config_guidepred['processing_mode'])
	query_index = list(dynamic_config_guidepred['query_index'])
	root_dir = str(dynamic_config_guidepred['output_directory'])
	# === mEdit offtarget prediction will process reference genome and alternate genomes ===
	# == Set internal variables for Off-target processing
	offtarget_genomes = offtarget_mode_formatting(mode, dynamic_config_guidepred)
	reference_genome = str(dynamic_config_guidepred['sequence_id'][0])

	# === Adjust SMK run based on processing_mode
	if mode == 'fast':
		allowed_rules = ['--omit-from symlink_genomes']

	# == Setup core off-target variables
	editors_list = []
	guides_per_editor_path = ''
	for index in query_index:
		for offtarget_genome, genome_type in offtarget_genomes:
			# == Set output paths ==
			guides_per_editor_path = str(
				f"{root_dir}/{mode}/jobs/{run_name}/guide_prediction-{reference_genome}/offtarget_prediction/{offtarget_genome}/{index}_")

			if genome_type == 'main_ref':
				# == Recover Guide Prediction filepath ==
				guides_report_path = Path(f"{root_dir}/{mode}/jobs/{run_name}/"
										  f"guide_prediction-{reference_genome}/guides_report_ref/{index}_Guides_found.csv")
				# TODO: Add BE report
				# == Group guides by editor and export DF as pickles by editor
				grouped_guide_dict = group_guide_table(guides_report_path, editing_tool_request)
				editors_list.extend(export_guides_by_editor(grouped_guide_dict, guides_per_editor_path))

			# Account for alternate genomes
			if genome_type == 'extended':
				guides_diff_path = Path(f"{root_dir}/{mode}/jobs/{run_name}/"
										f"guide_prediction-{reference_genome}/guides_report_{offtarget_genome}/{index}_Guide_differences.csv")
				grouped_diff_guide_dict = group_guide_table(guides_diff_path, editing_tool_request)
				editors_list.extend(export_guides_by_editor(grouped_diff_guide_dict, guides_per_editor_path))

	# === Export Variables to Configuration File ===
	# NOTE: This whole block (until the end) was inside the index loop
	config_template['guides_per_editor_path'] = guides_per_editor_path
	config_template['editors_list'] = list(set(editors_list))
	config_template['tmp_processing_casoff'] = f"{root_dir}/{config_template['tmp_processing_casoff']}"
	config_template['offtarget_genomes'] = {str(tup[0]): str(tup[1]) for tup in offtarget_genomes}
	config_template['offtarget_extended'] = {str(tup[0]): str(tup[1]) for tup in offtarget_genomes if
											tup[0] == 'extended'}
	config_template['reference_id'] = reference_genome

	# == Set the temporary directory up for CasOffinder ==
	set_export(config_template['tmp_casoff_path'])

	# === Write YAML configs to mEdit Root Directory ===
	write_yaml_to_file(config_template, dynamic_config_off_path)

	# === Invoke SMK Pipelines ===
	print("# == Calling Off-Target Prediction pipeline == #")
	for smk_setup_idx in range(len(allowed_rules)):
		try:
			# --> When cluster submission is switched on,
			launch_shell_cmd(f"snakemake "
							 f"--snakefile {project_file_path('smk.pipelines', 'offtarget_prediction.smk')} "
							 f"{smk_run_triggers} "
							 f"{allowed_rules[smk_setup_idx]} "
							 f"-j {ncores} "
							 f"{cluster_smk_setup} "
							 f"--configfile {config_db_path} "
							 f"{dynamic_config_guidepred_path} {dynamic_config_off_path} "
							 f"--use-conda "
							 f"--rerun-incomplete "
							 f"{dryrun_setup} ",
							 smk_verbosity[smk_setup_idx]
							 )
		except subprocess.CalledProcessError as e:
			print(f"Error: {e}")
		except ValueError:
			print(f"Process completed in a previous run. Moving to the next one...")



# == Native Modules ==
from copy import copy
from os.path import abspath
from pathlib import Path
import pickle
import subprocess
import sys
# == Installed Modules ==
import yaml
# == Project Modules ==
from prog.medit_lib import (compress_file,
							check_format,
							file_exists,
							handle_shell_exception,
							launch_shell_cmd,
							project_file_path,
							prRed,
							set_export,
							validate_editor_list,
							write_yaml_to_file)


def guide_prediction(args, jobtag):
	# === Load Run Parameters values ===
	# == Inputs and File tree setup
	query_input = args.query_input
	user_jobtag = args.user_jobtag
	root_dir = abspath(args.output)
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	# == Mode and associated values
	mode = args.mode
	mode_name = copy(mode)
	custom_vcfs = args.custom_vcf
	qtype = args.qtype_request
	# == Editor request and associated values
	editor_request = args.editor_request
	be_request = args.be_request
	cutdist = args.cutdist
	# == Custom editor required values
	pam = args.pam
	guide_length = args.guide_length
	pam_is_first = args.pam_is_first
	target_base = args.target_base
	result_base = args.result_base
	dsb_position = args.dsb_position
	editing_window = args.editing_window

	# === Load SLURM-related values ===
	ncores = args.ncores
	maxtime = args.maxtime
	parallel_processes = args.parallel_processes
	dry_run = args.dry_run

	# === Define dynamic SMK call variables ===
	allowed_rules = ['']
	cluster_smk_setup = ['']
	smk_verbosity = [True]
	smk_run_triggers = ''
	dryrun_setup = ''

	# ->=== OUTPUT SETUP ===<-
	# == Set import paths tied to the SMK pipeline
	config_db_path = f"{db_path_full}/config_db/config_db.yaml"
	# == Set export paths tied to the SMK pipeline ==
	config_dir_path = f"{root_dir}/config"
	# == Set export paths for Input Query Files
	query_input_path = f"{root_dir}/queries"
	# == Set export paths for dynamic YAML files ==
	dynamic_config_path = f"{config_dir_path}/config_{jobtag}.yaml"
	dynamic_cluster_path = f"{config_dir_path}/cluster_{jobtag}.yaml"
	# == Create sub-folders to host VCFs, and config files ==
	set_export(config_dir_path)

	# ->=== CONFIG FILES IMPORT ===<-
	#   == Load template configuration files ==
	config_template_path = project_file_path("smk.config", "medit_guide_pred.yaml")
	cluster_template_path = project_file_path("smk.config", "medit_cluster.yaml")
	with open(config_template_path, 'r') as config_handle:
		config_template = yaml.safe_load(config_handle)
	with open(cluster_template_path, 'r') as cluster_handle:
		cluster_template = yaml.safe_load(cluster_handle)
	#   == Load Database configuration file generated by db_set ==
	try:
		with open(config_db_path, 'r') as config_db_handle:
			config_db = yaml.safe_load(config_db_handle)
	except FileNotFoundError:
		handle_shell_exception("ConfigNotFound", config_db_path, True)
		exit(0)

	# ->=== INPUT CHECKS ===<-
	# == Check the input paths and copy them over to mEdits Filetree
	query_input_list = query_input.split(',')
	formatted_query_input_list = []
	query_index_count = 0
	# == Check file integrity
	for query_filename in query_input_list:
		if not file_exists(query_filename):
			handle_shell_exception(str('FileNotFound'), str(query_filename), True)
			exit(0)
		# == Copy query files in the right naming format for mEdit
		if mode_name == 'fast':
			mode_name = 'standard'
		formatted_query_filename = f"{query_input_path}/{mode_name}_{jobtag}_{query_index_count}.csv"
		formatted_query_input_list.append(formatted_query_filename)
		query_index_count += 1
		set_export(query_input_path)
		#   => Create a copy of the VCF in the internal mEdit directory
		launch_shell_cmd(f"cp {query_filename} {formatted_query_filename}", True)
	#   == Check the presence of custom vcf genome among the inputs ==
	if custom_vcfs:
		mode = 'vcf'
		custom_vcfs = args.custom_vcf.split(",")
	#   == Check the request to run on a cluster
	if parallel_processes:
		# --> Upon SLURM run request, the guide_prediction.smk is split in two separate runs
		# --> That's because samtool's conda package crashes on a libcrypto error when
		#       it's deployed by snakemake on a SLURM node
		cluster_smk_setup = ['', '--cluster "sbatch -t {cluster.time} -n {cluster.cores}" '
								 '--cluster-config config/medit_cluster.yaml']
		allowed_rules = ['--until "consensus_fasta"', '']
		smk_verbosity = [False, True]
	#   == Check the dry run request
	if dry_run:
		dryrun_setup = '-n'
	if user_jobtag:
		smk_run_triggers = '--rerun-triggers "mtime"'
	# == Check lists of editors and validate entries
	if editor_request not in ['clinical', 'custom']:
		with open(str(config_db["editors"]), 'rb') as editors_pkl_handle:
			editors_dict = pickle.load(editors_pkl_handle)
		editor_request = validate_editor_list(editor_request, list(editors_dict['all'].keys()), '--editor')
	if be_request not in ['default', 'custom']:
		with open(str(config_db["base_editors"]), 'rb') as be_pkl_handle:
			base_editors_dict = pickle.load(be_pkl_handle)
		be_request = validate_editor_list(be_request, list(base_editors_dict['all'].keys()), '--be')

	# == Check custom editors parameters
	if editor_request == 'custom':
		pam = check_format(pam, str, 'pam', 'XXX')
		pam_is_first = check_format(pam_is_first, bool, 'pamisfirst', None)
		guide_length = check_format(guide_length, int, 'guidelen', -1)
		dsb_position = check_format(dsb_position, int, 'dsb_pos', -10000)
	elif be_request == 'custom':
		pam = check_format(pam, str, 'pam', 'XXX')
		pam_is_first = check_format(pam_is_first, bool, 'pamisfirst', None)
		guide_length = check_format(guide_length, int, 'guidelen', -1)
		editing_window = check_format(editing_window, tuple, 'edit_win', (0, 0))
		target_base = check_format(target_base, str, 'target_base', 'X')
		result_base = check_format(result_base, str, 'result_base', 'X')

	# ->=== CHECK RUN MODE ===<-
	if mode == 'vcf':
		# == Enforce presence of custom vcf genome in this mode ==
		if not custom_vcfs:
			print("Please provide a VCF input file to run mEdit's vcf mode")
			sys.exit(1)
		# == Create a custom VCF directory when a vcf run is issued
		vcf_dir_path = f"{config_db['meditdb_path']}/{mode}/source_vcfs"
		set_export(vcf_dir_path)
		# == Check the config_db file for instructions on which human genome version should be used
		latest_reference = bool(config_db["latest_reference"])
		custom_reference = bool(config_db["custom_reference"])
		if latest_reference:
			config_template['sequence_id'] = ["latest_hg38"]
		if custom_reference:
			config_template['sequence_id'] = ["custom_reference"]
		# == VCF ID adjustment for custom vcf run ==
		#   => Import VCF file prefix information to config file
		tagged_genomes = []
		count_tag = 1
		for custom_vcf in custom_vcfs:
			loop_tag = f"{jobtag}_{count_tag}"
			vcf_filename = f"{loop_tag}.vcf"
			tagged_genomes.append(loop_tag)
			#   => Avoid re-creating custom VCF mirrors within the medit DB
			if not file_exists(f"{vcf_dir_path}/{vcf_filename}"):
				if not file_exists(f"{vcf_dir_path}/{vcf_filename}.gz"):
					#   => Create a copy of the VCF in the internal mEdit directory
					launch_shell_cmd(f"cp {custom_vcf} {vcf_dir_path}/{vcf_filename}", True)
					#   => Check VCF file compression and compress if necessary
					compress_file(f"{vcf_dir_path}/{vcf_filename}")
			count_tag += 1
		#   => Add any amount of custom vcf genomes to the config file
		config_template["vcf_id"] = tagged_genomes
	elif mode == 'fast':
		allowed_rules = ['--until "predict_guides" --omit-from consensus_fasta']
		# == 'fast' mode is a sub-mode of 'standard';
		# downstram processes must acknowledge that this is a standard run
		mode = 'standard'

	# === Assign Variables to Configuration File ===
	config_template['run_name'] = f"{mode}_{jobtag}"
	config_template['support_tables'] = db_path_full
	config_template['processing_mode'] = mode
	config_template['output_directory'] = root_dir
	config_template['variant_query'] = list(formatted_query_input_list)
	config_template['query_index'] = list(range(len(formatted_query_input_list)))
	# Assign run parameters to config
	config_template['qtype'] = qtype
	config_template['editor_request'] = editor_request
	config_template['be_request'] = be_request
	config_template['distance_from_cutsite'] = cutdist
	# Assign custom editor parameters to config
	config_template['pam'] = pam
	config_template['guide_length'] = guide_length
	config_template['pam_is_first'] = pam_is_first
	config_template['dsb_position'] = dsb_position
	config_template['editing_window'] = editing_window
	config_template['target_base'] = target_base
	config_template['result_base'] = result_base
	# Assign cluster options
	cluster_template['__default__']['cores'] = ncores
	cluster_template['__default__']['time'] = maxtime

	# === Write YAML configs to mEdit Root Directory ===
	write_yaml_to_file(config_template, dynamic_config_path)
	write_yaml_to_file(cluster_template, dynamic_cluster_path)

	# === Invoke SMK Pipelines ===

	print("# == Calling Guide Prediction pipeline == #")
	for smk_setup_idx in range(len(allowed_rules)):
		try:
			# --> When cluster submission is switched on,
			launch_shell_cmd(f"snakemake "
							 f"--snakefile {project_file_path('smk.pipelines', 'guide_prediction.smk')} "
							 # f"--directory {project_file_path('smk.pipelines', 'guide_prediction.smk')} "
							 f"-j {ncores} "
							 f"{smk_run_triggers} "
							 f"{allowed_rules[smk_setup_idx]} "
							 f"{cluster_smk_setup[smk_setup_idx]} "
							 f"--configfile {config_db_path} "
							 f"{dynamic_config_path} "
							 f"--use-conda "
							 f"--keep-going "
							 f"--rerun-incomplete "
							 f"{dryrun_setup}",
							 smk_verbosity[smk_setup_idx]
							 )
		except subprocess.CalledProcessError as e:
			print(f"Error: {e}")
		except ValueError:
			print(f"Process completed in a previous run. Moving to the next one...")



# == Native Modules
# read the contents of your README file
from pathlib import Path
# == Installed Modules
from setuptools import setup, find_packages
# == Project Modules
# == Readme setup
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='meditability',
    version='0.3.1',
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Daniel Bellieny, Taylor Hudson',
    author_email='bellieny@berkeley.edu, thudson@berkeley.edu',
    entry_points={
        "console_scripts": [
            "medit = prog:main",
            "ncbi_cross_db = prog:cross_db"
        ]
    },
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"smk.config": ["*.yaml"],
                  "smk.envs": ["*.yaml"],
                  "smk.pipelines": ["*.smk"]},
    include_package_data=True,
    install_requires=[
        'snakemake>=7.32.4',
        'pulp<2.8.0',
        'wrapt>=1.15.0',
        'biopython>=1.81',
        'pyyaml>=6.0',
        'importlib-resources>=6.1.1',
        'pytz>=2023.3',
        'boto3>=1.28.57',
        'alive-progress>=3.1.5',
        'xmltodict>=0.13.0'
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
    ],
)



# == Native Modules
import pickle
from os.path import abspath, isdir
# == Installed Modules
import yaml
# == Project Modules
from prog.medit_lib import file_exists


def ls_editors(args):
	# == Load Run Parameters values ==
	# print_editors = args.editors
	# print_base_editors = args.base_editors

	# === Load Database Path ===
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	config_db_path = f"{db_path_full}/config_db/config_db.yaml"

	if not file_exists(db_path_full):
		print("The database path directory could not be found.")
		exit(0)

	# === Load configuration file ===
	with open(config_db_path, 'r') as config_handle:
		config_db_obj = yaml.safe_load(config_handle)

	# === Load Editors Lists From Path Specified on config_db.yaml ===
	# if print_editors:
	with open(str(config_db_obj["editors"]), 'rb') as editors_pkl_handle:
		editors_dict = pickle.load(editors_pkl_handle)
		print(f"Available endonuclease editors: ")
		for k,v in editors_dict['all'].items():
			print("-----------------------------")
			print(f"name: {k}")
			print(f"pam, pam_is_first: {v[0]}, {v[1]}")
			print(f"guide_len: {str(v[2])}")
			if v[1] == True:# show both staggered positions for 5' pam editors
				print(f"dsb_position: {v[3]},{v[3]+5}")
			else:
				print(f"dsb_position: {v[3]}")
			if len(v) > 4:
				print(f"notes: {v[4]}")
			print(f"5'-{v[0] if v[1] else ''}{'x' * int(v[2])}{'' if v[1] else v[0]}-3'")


	# if print_base_editors:
	with open(str(config_db_obj["base_editors"]), 'rb') as be_pkl_handle:
		base_editors_dict = pickle.load(be_pkl_handle)
		print(f"Available base editors: ")
		for k,v in base_editors_dict['all'].items():

			print("-----------------------------")
			print(f"name: {k}")
			print(f"pam, pam_is_first: {v[0][0]}, {v[0][1]}")
			print(f"guide_len: {str(v[0][2])}")
			print(f"edit_win: {str(v[0][3])[1:-1]}")
			print(f'target_base, result_base: {v[1][0][0]} ---> {v[1][0][1]}')
			if len(v[0]) > 4:
				print(f"notes: {v[0][4]}")
			print(f"5'-{v[0][0] if v[0][1] else ''}{'x' * int(v[0][2])}{'' if v[0][1] else v[0][0]}-3'")




# Native Modules
import pickle
import os
import re
# Installed Modules
import pandas as pd
# Project Modules
from dataH import DataHandler
from annotate import Transcript


###############
# Main Script with Fetch_Guides Class for running pipeline
###############
def set_export(outdir):
	# Create outdir if inexistent
	if not os.path.exists(outdir):
		os.makedirs(outdir)
	return outdir


class Fetch_Guides:

	def __init__(self,
				 queries: list,
				 qtype: str,
				 editor_request: list,
				 be_request: list,
				 editors: dict,
				 base_editors: dict,
				 dist_from_cutsite: int,
				 datadir: str,
				 fasta_path: str,
				 annote_path: str,
				 **kwargs):
		"""
		:param queries: list of query terms, either in hgvs format - 'NM_000518.5:c.114G>A' or coords 'chr11:5226778C>T'
		(COORDS ALLELES MUST BE PLUS STRAND!!)
		:param qtype: 'hgvs' or 'coord'
		  --> if 'hgvs', providing the coordinates in the kwargs with 'hgvscoord' can reduce processing time
		  --> hgvs assumes the query is already in clinvar and will generate a variant report with the gene report,
		  --> if 'coord' then just gene report is created
		:param editor_request: 'clinical', 'custom', name (list from editor choices)
			--> custom must contain kwargs - pam, pam_is_first,guide_length, dsb_position
		:param be_request: 'off','default',custom, or select BE editor for base editor choices below
			--> custom must contain kwargs - pam, pam_is_first, guide_length, editing_window, target_base,result_base
		:param editors: Dictionary containing information on the current set of editors supported by mEdit
		:param base_editors: Dictionary containing information on the current set of base editors supported by mEdit
		:param dist_from_cutsite: this is how farway the cutsite/dsb_position may be from the start of the variant position
		:param datadir: folder where tables and pre-computed data live
		:param fasta_path: *Unsure using chromsome seperate files right now but unsure if this will be permenant
		:param kwargs:
		"""
		##-----------------User Required_Inputs--------------------##
		if qtype == 'hgvs':
			self.queries = self.validate_hgvs(queries)
		if qtype == 'coord':
			self.queries = self.validate_coord(queries)
		self.qtype = qtype
		self.editor_request = editor_request  ## clinical, custom, list
		self.be_request = be_request ## off, default, custom, list

		##------------Optional inputs-------------###
		self.dist_from_cutsite, self.window = self.validate_dist_from_cutsite(dist_from_cutsite)

		##----endonuclease and BE search params
		self.editor_lib = editors
		self.be_lib = base_editors
		self.guide_length = 20
		self.pam =str()
		self.pam_is_first = False
		self.dsb_position = int()
		self.editing_window = [4,8]
		self.conversion = str()
		self.search_params = self.configure_search_params(kwargs)
		self.BE_search_params = self.configure_BE_params(kwargs)


		# input paths/folders
		self.processed_tables = f"{datadir}/processed_tables"  # folder with cleaned clinvar/hpa tabs
		self.HGVSlookup_path = f"{self.processed_tables}/HGVSlookup.csv"
		self.fasta_path = fasta_path
		self.annote_path = annote_path

		##---------------Outputs--------------------##
		self.snv_info = {}  # {chrom: (queryterm,tid,eid,gene,strand,ref,alt,feature,extracted_seq,rf,coord)}
		self.all_guides = {}
		self.all_BE = {}
		self.not_found = {}
		self.all_variant = pd.DataFrame()
		self.all_gene = pd.DataFrame()
		self.nguides_df = pd.DataFrame()


	def configure_search_params(self,kwargs):
		"""
		set paramteres for the selected editor or editors(not BE editors)
		"""
		search_params = {}
		# search for all guides
		if self.editor_request == str('clinical'):
			search_params = self.editor_lib['clinical']
		# set custom editor params
		elif self.editor_request == str('custom'):
			search_params = self.set_custom_params(kwargs)

		# search for selected subset
		else:
			# === At this point, the user-defined list has been thoroughly validated by guide_prediction.py ===
			print(f"*********** EDITOR REQUEST: {self.editor_request} -- Type: {type(self.editor_request)}")
			for editor in list(self.editor_request):
				search_params[editor] = self.editor_lib['all'][editor]

		print(f'Editor(s) set to: {[x for x in search_params.keys()]}')
		return search_params

	def set_custom_params(self,kwargs):
		try:
			self.pam = kwargs['pam'].upper()
		except KeyError:
			print("custom editor selection MUST contain pam argument")

		try:
			self.pam_is_first = kwargs['pam_is_first']
			if not isinstance(kwargs['pam_is_first'], bool):
				raise TypeError(f"'pam_is_first' must be of type 'bool', got {type(kwargs['pam_is_first'])} instead.")
		except KeyError:
			print("custom editor selection MUST also have  pam_is_first in kwargs")

		try:
			self.guide_length= kwargs['guide_length']
			if not isinstance(kwargs['guide_length'], int):
				raise TypeError(f"'guide_length' must be of type 'int', got {kwargs['guide_length'].__name__} instead.")
		except KeyError:
			print("custom editor selection MUST also have guide_length in kwargs")
			print(f"custom editor guide length being set to {self.guide_length}")
			pass
		try:
			dsb_position = str(kwargs['dsb_position']).split(",")
			self.dsb_position = int(dsb_position[0])
		except KeyError:
			print("custom editor selection MUST also have dsb_loc in kwargs")

		params = {'custom_endonuclease': (self.pam, self.pam_is_first, self.dsb_position, self.guide_length, '')}

		print(f'Your custom editor has a {"5 prime PAM" if self.pam_is_first else "3 prime PAM"} set to {self.pam}')
		print(f'with a spacer length of {self.guide_length}' )
		print(f"5'-{self.pam if self.pam_is_first else ''}{'x'*self.guide_length}{'' if self.pam_is_first else self.pam}-3'")
		return params

	def configure_BE_params(self,kwargs):
		# sets base editor search params, each key is a list of 2 or more; refernce seq search params,
		# then any set that follows starts with the conversion (ex. 'AG' is A --> G) and then the base editors that have the same params

		be_search_params = {}
		if self.be_request == 'default':
			be_search_params = self.be_lib['default']
		elif self.be_request == 'off':
			be_search_params = {}
		elif self.be_request == 'custom':
			be_search_params = self.set_be_custom_params(kwargs)
		else:
			# === At this point, the user-defined list has been thoroughly validated by guide_prediction.py ===
			for base_editor in list(self.be_request):
				be_search_params[base_editor] = self.be_lib['all'][base_editor]

		print(f'Base Editor(s) set to: {[x for x in be_search_params.keys()]}')
		return be_search_params

	def set_be_custom_params(self, kwargs):
		try:
			self.pam = kwargs['pam'].upper()
		except KeyError:
			print("custom editor selection MUST contain pam argument")

		try:
			self.pam_is_first = kwargs['pam_is_first']
			if not isinstance(kwargs['pam_is_first'], bool):
				raise TypeError(f"'pam_is_first' must be of type 'bool', got {type(kwargs['pam_is_first'])} instead.")
		except KeyError:
			print("custom editor selection MUST also have  pam_is_first in kwargs")
			print(f"custom editor pam_is_first is being set to {self.pam_is_first}")
			pass

		try:
			self.guide_length= int(kwargs['guide_length'])
		except KeyError:
			print("custom editor selection MUST also have guide_length in kwargs")
			print(f"custom editor pam_is_first is being set to {self.guide_length}")
			pass
		try:
			target_base = kwargs['target_base'].upper()
			result_base = kwargs['result_base'].upper()
			if target_base not in ['A','G','C','T'] or target_base not in ['A','G','C','T']:
				raise TypeError("target_base or result_base need to equal 'A','C','G' or 'T'")
			self.conversion = target_base + result_base
		except KeyError:
			print("custom editor selection MUST also have 'target_base' and 'result_base' in kwargs")

		params = {'custom_be': [(self.pam, self.pam_is_first,  self.guide_length, self.editing_window, ''),(self.conversion , 'custom_be')]}

		print(f"Your custom base editor has a {'5 prime PAM' if self.pam_is_first else '3 prime PAM'} set to {self.pam}")
		print(f'with a spacer length of {self.guide_length}')
		print(f"The base editor window is between position {self.editing_window[0]} and {self.editing_window[1]}")
		print(f"and will convert {self.conversion[0]}---->{self.conversion[1]}")
		print(f"5'-{self.pam if self.pam_is_first else ''}{'x'*self.guide_length}{'' if self.pam_is_first else self.pam}-3'")
		return params


	def validate_dist_from_cutsite(self,dist_from_cutsite):
		'''
		Makes sure the extracted genome window is big enough
		to accomodate distance from cutsite
		'''
		if dist_from_cutsite <=7:
			return dist_from_cutsite, 50
		elif dist_from_cutsite > 7 and dist_from_cutsite <=200:
			return dist_from_cutsite, 50 + (dist_from_cutsite - 7)
		else:
			print('!!! The dist_from_cutsite must be under 200 !!!')
			print(f'The current dist_from_cutsite = {self.dist_from_cutsite} is being changed  to 200')
			return 200,50 + (200 - 7)
			

	def write_besearch_params(self, outfile):
		# writes pickle of selected be guide search params for later use in process_genome
		with open(outfile, 'ab') as gfile:
			pickle.dump(self.BE_search_params, gfile)


	def write_gsearch_params(self, outfile):
		# writes pickle of selected guide search params for later use in process_genome
		# 'editor', 'pam', '5prime_pam','guide_length', 'DSB site', 'notes'
		with open(outfile, 'ab') as gfile:
			pickle.dump(self.search_params, gfile)


	def write_snv_site_info(self, outfile):
		'''
		#writes pickle of SNV site info for later use in process genome
		#query, tid, eid, strand, ref, alt, feature_annotation, extracted_seq, codons, coord
		'''
		with open(outfile, 'ab') as sfile:
			pickle.dump(self.snv_info, sfile)

	def write_not_found(self, outfile):
		pd.Series(data=self.not_found.values(), index=self.not_found.keys()).to_csv(outfile)


	def write_guide_csv(self, guides,outfile):
		df = pd.DataFrame(guides)
		df['Guide_ID'] = [y + str(x) for x, y in zip(list(df.index), list(df['Guide_ID']))]
		df.to_csv(outfile, index=False)
		return df

	def add_clinvar(self, gadf):
		self.all_variant = pd.concat([self.all_variant, gadf])

	def write_reports(self, gene_out, variant_out,nguides_out):
		guides_found = True if len(self.all_guides['QueryTerm']) >0 else False

		# writes
		all_tids = []
		searched_queries = []
		for chrom in self.snv_info.keys():
			for v in self.snv_info[chrom]:
				all_tids.append(v[1])
				searched_queries.append(v[0])

		#creates gene table from queries
		tempgene = pd.read_csv(f"{self.processed_tables}/gene_tables/gene_tables.csv.gz")
		self.all_gene = tempgene.loc[tempgene['TranscriptID'].isin(list(all_tids))]
		self.all_gene.to_csv(gene_out, index=False)

		#tallies number of guides per query
		if guides_found:
			allgdf = pd.DataFrame(self.all_guides)
			nguides_per_query = allgdf.groupby(['QueryTerm']).size().reset_index()
			nguides_per_query = nguides_per_query.rename(columns={0: 'Number of Guides Found'})
			self.nguides_df = pd.DataFrame(searched_queries,columns = ['QueryTerm']).join(nguides_per_query.set_index('QueryTerm'), on='QueryTerm')
			self.nguides_df['Number of Guides Found'] = self.nguides_df['Number of Guides Found'].fillna(0).astype('int')
			self.nguides_df.to_csv(nguides_out, index=False)

		if self.qtype == 'hgvs':
			self.all_variant.to_csv(variant_out, index=False)


	def add_not_found(self, nfqueries, reason):
		# adds queries that are not found ex: no guides or hgvs not found
		for nf in nfqueries:
			self.not_found[nf] = reason

	@staticmethod
	def extract_seqs(searchseq, pos, ref, alt, window):
		# extracts the sequence +/windowbp surrounding a SNV then swaps ref for alt allele

		alt = alt.lower()
		searchseq = searchseq.upper()

		if len(ref) == len(alt):  ##substitution
			extracted_seq = str(searchseq[pos - window:pos + window])
			variant_seq = (extracted_seq[0:window] + alt + extracted_seq[window + len(alt):])#.upper()


		elif len(ref) > len(alt):  # deletion
			diff = len(ref) - len(alt)
			extracted_seq = str(searchseq[pos - window:(pos + window + diff)])
			variant_seq = (extracted_seq[0:window] + alt + extracted_seq[window + len(ref):])#.upper()

		elif len(ref) < len(alt):  # insertion
			extracted_seq = str(searchseq[pos - window:pos + window - (len(alt))])
			variant_seq = (extracted_seq[0:window] + alt + extracted_seq[window:])#.upper()
		# print(new_seq)

		else:
			print('VARIANT REF AND ALT DO NOT COMPLY')
			print(ref, alt, pos)
		return variant_seq

	def get_chroms(self, queries):
		# finds unique chromosome for each hgvs
		# needs to happen in order to select right fasta file
		hgvs_tab = pd.read_csv(self.HGVSlookup_path)
		q_prefixes = [x.split(':')[0].split('.')[0] for x in queries]
		chroms = set(hgvs_tab.loc[hgvs_tab['HGVS_ID'].isin(q_prefixes), 'Chromosome'])
		return chroms

	def fetch_query_info(self):
		# Gets Transcript info
		# If quering by HGVSID with no other info then need to get chromsome/location/alt/ref
		if self.qtype == 'hgvs':
			print("Looking up HGVS in Clinvar.......")
			chroms = self.get_chroms(self.queries)

			for ch in chroms:
				df = pd.read_csv(f"{self.processed_tables}/variant_tables/{ch}_variant.txt",
								 low_memory=False)
				gadf = df.loc[df['HGVS_Simple'].isin(self.queries)]
				self.add_clinvar(gadf)
				self.snv_info[ch] = gadf[['HGVS_Simple', 'PositionVCF', 'RefAlleleVCF', 'AltAlleleVCF']].to_dict('tight')['data']

			#record not found
			found = []
			for k in self.snv_info.keys():
				for v in self.snv_info[k]:
					found.append(v[0])
			notfound = list(set(self.queries).difference(set(found)))
			self.add_not_found(notfound, 'hgvs not found in medit variant database')

		# Else All information is given to find transcript info
		else:
			coord_fmt = r'chr[0-9MTXY]*:(\d*)([ATCG]{1,10})\>([ATCG]{1,10})'
			for query in self.queries:
				ch = query.split(':')[0].replace('chr', '')
				if ch not in self.snv_info.keys():
					self.snv_info[ch] = []
				snvpos, alt, ref = list(re.search(coord_fmt, query).groups())
				self.snv_info[ch].append([query, int(snvpos), alt, ref])

	def find_transcript_info(self):

		print("Gathering Variant Genomic Info from RefSeq file.......")

		for ch, data in self.snv_info.items():  # find transcript info
			new_data = []
			snvcoords = [f'chr{ch}:{d[1]}-{d[1]}' for d in data]
			try:
				Transcript.load_transcripts(self.annote_path, snvcoords)
			except IndexError:
				# TODO: Something on commit f6f4b19 has changed the communication between fetchGuides and annotate.py
				# TODO: @ACTG802 --> please advise on whether this needs fixing or just a bypass message would be enough
				print(f"WARNING: function 'load_transcripts' from annotate.py got the wrong number of indices from snvcoords")
				print(f"This is how it looks like: {snvcoords}")
				continue

			# === Transitioning SeqIO.read to direct import of Pickled SeqRecord Objects ===
			chr_fasta_path = f"{self.fasta_path}/chr{str(ch)}.pkl"
			try:
				print(f"Finding transcripts information on chromosome chr{str(ch)}")
				with open(chr_fasta_path, 'rb') as pfile:
					fasta = pickle.load(pfile)
			except FileNotFoundError:
				print(
					f"The file {chr_fasta_path} was not found. Please regenerate background data and check the target directory")
				continue
			except pickle.UnpicklingError:
				print(f"The file {chr_fasta_path} is not in the correct format. Please regenerate background data")
				continue

			for d in data:

				query, snvpos, ref, alt = d
				if re.search('[^ATCG]', ref + alt) is None:

					snvcoord = f'chr{ch}:{d[1]}-{d[1]}'
					tx = Transcript.transcript(snvcoord)

					if tx == 'intergenic':
						tid, eid, gname = '-', '-', '-'
						feature_annotation = tx
						rf, strand = 'None', '+'
						extracted_seq = self.extract_seqs(searchseq=fasta.seq, pos=snvpos, ref=ref, alt=alt,
														  window=self.window)


					else:
						eid, tid, gname, strand, txstart = tx.tx_info()
						tx_seq = tx.get_tx_seq(fasta)
						t_snvpos = int(snvpos) - int(txstart)
						extracted_seq = self.extract_seqs(searchseq=tx_seq, pos=t_snvpos - 1, ref=ref, alt=alt,
														  window=self.window)

						if len(extracted_seq) != self.window * 2:
							# if flanking or utr the extracted seq needs to come from the chromosome file
							print('Extracted Sequence Length OUT OF BOUNDS in Transcript Sequence ')
							print('Searching chromosome sequence instead')

							extracted_seq = self.extract_seqs(
								searchseq=fasta.seq[snvpos - self.window * 3:snvpos + self.window * 3],
								pos=self.window * 3,
								ref=ref,
								alt=alt,
								window=self.window)

						feature_annotation, rf = tx.feature, tx.rf

					new_data.append(
						[query, tid, eid, gname, strand, ref, alt, feature_annotation, extracted_seq, rf,
						 f"chr{str(ch)}:{str(snvpos)}"])

				else:
					self.add_not_found(query, 'ref or alt allele contain non-ATCG characters')
			self.snv_info[ch] = new_data

	@staticmethod
	def validate_hgvs(queries):
		'''
		standardizes input hgvs and checks formating
		'''
		rprefix = r"((N(M|G|C|R)_[\d.]*)|(m))"
		rsuffix = r"(:(c|m|g|n)\S*)"
		validated_queries = []
		for q in set(queries):
			if re.search(rsuffix, q) and re.search(rprefix, q):
				validated_queries.append(re.search(rprefix, q).groups()[0] + re.search(rsuffix, q).groups()[0])
		n = len(validated_queries)
		print(f'{n} out of {len(queries)} HGVS IDs validated')
		if n == 0:
			print('Query are not in the correct HGVS Format')
		return validated_queries

	@staticmethod
	def validate_coord(queries):
		'''
		standardizes input coordinate and checks formatting
		'''
		# q = 'chr11:5226778C>T'
		# coord_fmt = r'(chr[0-9XYM]*:\d*(A|T|C|G)>(A|T|C|G))'
		coord_fmt = r'(chr[0-9XYM]*:\d*([ATCG]{1,10})>([ATCG]{1,10}))'
		validated_queries = []
		for q in set(queries):
			if re.search(coord_fmt, q):
				validated_queries.append(re.search(coord_fmt, q).groups()[0])
			else:
				print(q)
		n = len(validated_queries)
		print(f'{n} out of {len(queries)} Coordinates IDs validated')
		if n == 0:
			print('Query are not in the correct Coordinate + allele Format')
		return validated_queries

	def run_FetchGuides(self, outfile_guides, outfile_be_guides, models_dir):
		#global dh, query
		self.fetch_query_info()
		self.find_transcript_info()
		print('Finding Guides')
		for ch, data in self.snv_info.items():

			for d in data:
				try:
					query, tid, eid, gname, strand, ref, alt, feature_annotation, extracted_seq, codons, coord = d
				except ValueError:
					print(
						f"WARNING: fetchGuides.py --> The query below has the wrong number of values to unpack. Needs further investigation:\n{d}")
					continue
				dh = DataHandler(query, strand, ref, alt, feature_annotation, models_dir, extracted_seq, codons, coord,
								 gname, self.dist_from_cutsite)

				guides, beguides = dh.get_Guides(self.search_params, self.BE_search_params)

				print(f"{len(guides['gRNA'])+len(beguides['gRNA'])} found for {query}")

				if len(guides['gRNA']) > 0:
					if len(self.all_guides.keys()) == 0:
						self.all_guides = guides
					else:
						for k, v in guides.items():
							self.all_guides[k] += v

				if len(beguides['gRNA']) > 0:
					if len(self.all_BE.keys()) == 0:
						self.all_BE = beguides
					else:
						for k, v in beguides.items():
							self.all_BE[k] += v

				if len(guides['gRNA']) +  len(beguides['gRNA']) ==0:
					self.add_not_found([query], 'no guides found')

		if len(self.all_guides.keys()) != 0:
			self.write_guide_csv(self.all_guides,outfile_guides)
		else:
			print('No Endonuclease Guides found for any queries')


		if len(self.all_BE.keys()) != 0:
			self.write_guide_csv(self.all_BE, outfile_be_guides)
		else:
			if self.be_request != 'off':
				print('No Base Editor Guides found for any queries')


def main():
	# SNAKEMAKE IMPORTS
	# === Inputs ===
	input_file = str(snakemake.input.query_manifest)
	fasta_path = str(snakemake.input.assembly_dir_path)
	# === Outputs ===
	guides_report = str(snakemake.output.guides_report_out)
	gene_report = str(snakemake.output.gene_report)
	variant_report = str(snakemake.output.variant_report)
	nguides_report = str(snakemake.output.nguides_report)
	be_report = str(snakemake.output.be_report_out)
	guide_search_params_path = str(snakemake.output.guide_search_params)
	guide_be_search_params_path = str(snakemake.output.guide_be_search_params)
	snv_site_info_path = str(snakemake.output.snv_site_info)
	guides_not_found_path = str(snakemake.output.guides_not_found_out)
	# === Params ===
	resultsfolder = set_export(str(snakemake.params.main_out))
	#   == Processed tables branch ==
	datadir = str(snakemake.params.support_tables)
	annote_path = str(snakemake.params.annote_path)
	# == Editor Parameters
	editors_path = str(snakemake.params.editors)
	base_editors_path = str(snakemake.params.base_editors)
	models_path = str(snakemake.params.models_path)
	distance_from_cutsite = int(snakemake.params.distance_from_cutsite)
	#   == Run Parameters ==
	qtype = str(snakemake.params.qtype)
	be_request = snakemake.params.be_request
	editor_request = snakemake.params.editor_request
	# 	== Custom Editor Parameters ==
	pam = str(snakemake.params.pam)
	guide_length = int(snakemake.params.guide_length)
	pam_is_first = bool(snakemake.params.pam_is_first)
	dsb_position = int(snakemake.params.dsb_position)
	editing_window = tuple(snakemake.params.editing_window)
	target_base = str(snakemake.params.target_base)
	result_base = str(snakemake.params.result_base)

	# == Create dummy files for optional outputs
	optional_outputs = [be_report, variant_report, gene_report, nguides_report]
	for report in optional_outputs:
		# Create an empty DataFrame
		df = pd.DataFrame()
		# Export the empty DataFrame to a CSV file
		df.to_csv(str(report))

	# == Input Setup ==
	df = pd.read_csv(input_file)
	queries = list(df.iloc[:, 0])

	# == Editors / BEs Setup ==
	with open(editors_path, 'rb') as edfile:
		editors = pickle.load(edfile)
	with open(base_editors_path, 'rb') as befile:
		base_editors = pickle.load(befile)

	# == Report processed input variables ==
	# print(f"""
	# Currently running fetchGuides.py
	# INPUT VARIABLES:
	# 	n of Queries: {len(queries)}
	# 	Query Type: {qtype}
	# 	be_request: {be_request}
	# 	editor_request: {editor_request}
	# PATH TO REFERENCE:
	# 	-> {fasta_path}
	# SUPPORT DATA DIRECTORY:
	# 	-> {datadir}
	# OUTPUTS TO:
	# 	--> {resultsfolder}
	# """)

	# == Get query items ==
	fg = Fetch_Guides(queries,
					  qtype,
					  editor_request,
					  be_request,
					  editors,
					  base_editors,
					  distance_from_cutsite,
					  datadir,
					  fasta_path,
					  annote_path,
					  pam=pam,
					  pam_is_first=pam_is_first,
					  guide_length=guide_length,
					  dsb_position=dsb_position,
					  editing_window=editing_window,
					  target_base=target_base,
					  result_base=result_base
					  )
	# == Set up object and run core methods ==
	fg.run_FetchGuides(guides_report, be_report, models_path)

	# == Export Intermediate files ==
	fg.write_snv_site_info(snv_site_info_path)
	fg.write_gsearch_params(guide_search_params_path)
	fg.write_besearch_params(guide_be_search_params_path)

	# == Export Variant, Guide Totals and Gene tables ==
	fg.write_reports(gene_report, variant_report, nguides_report)

	# == Export Not Found table
	fg.write_not_found(guides_not_found_path)


if __name__ == "__main__":
	main()



# Native Modules
import pickle
import subprocess
# Installed Modules
import pandas as pd
import vcf
# == Pysam REQUIRED ==
# Project Modules
from dataH import DataHandler

#################################
# Compares ALT VCF to Guide found in Hg38
# Searches and compares changes in either guide or pam
# If new guide is created --> submits query back to Fetch guides
#################################

def make_dict(ALTguides_dict, new_guides_set,new_beguides_set,set_header):

	if set_header:
		header = list(set(list(new_beguides_set.keys()) + list(new_guides_set.keys())))
		for k in header:
			ALTguides_dict[k] = []

	if len(new_guides_set['gRNA']) > 0:
		value_length = len(new_guides_set['gRNA'])
		for k in ALTguides_dict.keys():
			if k in new_guides_set.keys():
				ALTguides_dict[k] += [x for x in new_guides_set[k]]
			else:
				ALTguides_dict[k] += ["-"] * value_length

		if len(new_beguides_set['gRNA']) > 0:
			value_length = len(new_beguides_set['gRNA'])
			for k in ALTguides_dict.keys():
				if k in new_beguides_set.keys():
					ALTguides_dict[k] += [x for x in new_beguides_set[k]]
				else:
					ALTguides_dict[k] += ["-"] * value_length
		return ALTguides_dict


def extract_vcf_record(REFcoord, vcf_fname, window):
	'''
	searches for overlapping variant at the provided REF genome coordinate $ window
	'''
	records_found = []
	ch, pos = REFcoord.split(':')
	vcf_reader = vcf.Reader(filename = vcf_fname, compressed = True)
	for record in vcf_reader.fetch(ch, int(pos) - window, int(pos) + window):
		records_found.append(record)
	return records_found


def make_ALTseq(extracted_seq, pos, ref, alt):
	'''
	Given the Reference Genome sequence (variant incorporated), adds mutations from ALT VCF to the sequence
	:param extracted_seq: Ref sequence
	:param pos: Alt genome variant start pos
	:param ref: Reference Allele
	:param alt: Alt Genome alt allele
	'''
	variant_seq = ''
	if len(ref) == len(alt):  ##substitution
		variant_seq = (extracted_seq[0:pos] + alt.lower() + extracted_seq[pos + len(alt):])
	else:
		print("SNV sites only accepted at this time")
	return variant_seq


def extract_variant_info(record, REFextracted_seq, REFcoord):
	'''
	for a given alt record found determine the ALT information
	and create a new extracted sequence incorporating the ALT allele
	'''
	## RECORD INFO FOR SAMPLE =1
	ALTalt_list, ALTref =record.ALT,record.REF
	call = record.samples[0]
	zyg = 'heterozygous' if call.is_het else 'homozygous'
	window = int(len(REFextracted_seq)/2)

	## if there is more than 2 ALT alleles
	if len(ALTalt_list) == 2:
		zyg = zyg + '-biallelic'
	if len(ALTalt_list) > 2:
		zyg = zyg + '-multiallelic'

	## THE first alt allele info
	vtype = record.ALT[0].type
	ALTalt = ALTalt_list[0].sequence ### <-- this program only incorporates the first alternative allele
	REFstart = int(REFcoord.split(':')[1]) - window
	ALTstart, ALTend = record.start, record.end
	ALTcoord = f'{record.CHROM}:{record.POS}'
	rel_pos = (ALTstart -REFstart)+ 1
	ALTalts = ",".join([str(x) for x in record.samples[0].site.alleles][1:])
	ALTid = f'{ALTcoord}:{ALTref}>{ALTalts}'

	#print('alt, ref',record.ALT, record.REF)
	#print('relative position',rel_pos)
	#print('check extracted seq == ref',REFextracted_seq[rel_pos],record.REF)
	#print('ALT starts and stop',ALTstart,ALTend)

	#if len(record.REF) == len(alt): ## substitution
	#    vtype = vtype + '-sub'
	#    new_seq = REFextracted_seq[0:rel_pos[0]] + alt.sequence + REFextracted_seq[rel_pos[1]:]
		#print(new_seq)

	#elif len(record.REF) > len(alt): # deletion
	#    vt = vtype + '-del'
		#new_seq = hg38extracted_seq[0:rel_pos[0]] + alt.sequence + hg38extracted_seq[rel_pos[1]:]
	#    new_seq = 'undetermined'

	#elif len(record.REF) < len(alt): # insertion
	#    vt = vtype + '-ins'
	#    new_seq = REFextracted_seq[0:rel_pos[0]] + alt.sequence + REFextracted_seq[rel_pos[1]:]
	#    new_seq = new_seq[0:len(REFextracted_seq)-1]


	#else:
	#    new_seq = 'undetermined'

	return ALTref, ALTalt, ALTalts, ALTcoord,zyg, vtype, rel_pos, ALTid


def find_overlapping_variants(filtered_vcf, models_path, hg38_snvinfo, search_params,be_search_params):
	ALTguides_dict = {}
	ALTinfo = {}
	set_header = True

	for ch, data in hg38_snvinfo.items():
		for d in data:
			query, tid, eid, gname,strand, REFref, REFalt, feature_annotation, REFextracted_seq, codons, REFcoord = d

			# Check VCF if variant exists in hg38 extracted_seq
			window = len(REFextracted_seq)/2
			records_found = extract_vcf_record(REFcoord=REFcoord, vcf_fname = filtered_vcf, window=window)

			if len(records_found) > 0:

				ALTseq = REFextracted_seq

				for rec in records_found:
					ALTref, ALTalt, ALTalts, ALTcoord, zyg, vtype, rel_pos, ALTid= extract_variant_info(rec,str(REFextracted_seq),REFcoord)

					if rel_pos != window: #make sure the alt genome mutation is not the same as the one in clinvar or it will be duplicated
						ALTinfo[query] = {ALTid : [ALTid,ALTref, ALTalt, ALTalts, ALTcoord, zyg, vtype, window -rel_pos if rel_pos>window else -1*rel_pos]}
						ALTseq = make_ALTseq(extracted_seq = ALTseq, pos = rel_pos, ref = REFalt, alt = ALTalt)

				if ALTseq != REFextracted_seq:
					# find quides
					dh = DataHandler(query, strand, REFref, REFalt, feature_annotation, models_path, ALTseq, codons,
									 REFcoord, gname)
					new_guides_set, new_beguides_set  = dh.get_Guides(search_params,be_search_params)

					make_dict(ALTguides_dict, new_guides_set, new_beguides_set, set_header)
					set_header = False
	return ALTinfo, ALTguides_dict

def guide_compare(guides_report,be_report,ALTguides_dict,ALTinfo,refgenome_name,altgenome_name):
	'''
	Compares the guides generated from the ALT and REf genomes.
	Creates a DF that shows the ALT genomes impact on REF guides

	:param guides_report: Reference Genome guide CSV
	:param guides_report: Reference Genome base editor guide CSV
	:param ALTguides_dict: Guide found with variations found in ALT genome
	:param ALTinfo: ALT genome VCF variant into
	:param refgenome_name: example 'Hg38'
	:param altgenome_name: example 'H02557'
	'''

	new_guides = []

	altgdf = pd.DataFrame(ALTguides_dict)
	refgdf = pd.read_csv(guides_report) # combine both be guides and endo guides together
	refbedf = pd.read_csv(be_report)
	refgdf = pd.concat([refgdf,refbedf])
	altgdf =altgdf.loc[:,refgdf.columns] # reorder cols to match

	refgdf = refgdf[refgdf['QueryTerm'].isin(list(ALTinfo.keys()))]
	REFrows = refgdf.to_dict('tight')['data']
	ALTrows = altgdf.to_dict('tight')['data']

	rename_columns = [f'{altgenome_name} {x}' if 'core' in x or 'B' in x else x for x in altgdf.columns[8:]]
	columns = ['QueryTerm','GeneName', 'Editor',f'{altgenome_name} Guide Impact',
			   'Coordinates','Strand',f'{altgenome_name} gRNA', f'{altgenome_name} Pam',
			   f'{refgenome_name} gRNA', f'{refgenome_name} Pam'] + rename_columns


	for rrow in REFrows:
		query, gene,ed,gid,coord,strand, REFgrna, REFpam = rrow[0:8]

		cnt = 1
		for arow in ALTrows:
			if query == arow[0] and ed ==arow[2] :
				ALTgrna, ALTpam = arow[6:8]

				if [REFpam, REFgrna] == [ALTpam, ALTgrna]:  # unchanged
					#print('impact;unchanged', REFpam, '->', ALTpam, REFgrna, '->', ALTgrna)
					ALTrows.remove(arow)
					cnt -= 1
					break

				elif REFpam == ALTpam and ALTgrna != REFgrna:  # same pam which means change is in grna
					newrow = rrow[0:3] + ['protospacer changed & conserved'] + rrow[4:6] + arow[6:8] + rrow[6:8] + arow[8:]
					new_guides.append(newrow)
					#print('impact;grna_changed_conserved', REFpam, '->', ALTpam, REFgrna, '->', ALTgrna)
					ALTrows.remove(arow)
					cnt -= 1
					break

				elif REFgrna == ALTgrna:  # pam is changed but conserved
					newrow = rrow[0:3] + ['PAM changed & conserved'] + rrow[4:6] + arow[6:8] + rrow[6:8] + arow[8:]
					new_guides.append(newrow)
					#print('impact;pam_changed_conserved', REFpam, '->', ALTpam, REFgrna, '->', ALTgrna)
					ALTrows.remove(arow)
					cnt -= 1
					break
				else:
					pass

		if cnt == 1:  # old guides remaining and not matched == no longer exsist
			newrow = rrow[0:3] + ['PAM changed & removed'] + rrow[4:6] +["-","-"] + rrow[6:8] + len(rrow[8:]) * ['-']
			new_guides.append(newrow)
			#print('impact;pam_changed_removed', REFpam, '->', '-', REFgrna, '->', '-')


	if len(ALTrows) > 0:  # new guides remaining and not matched == new guides are made
		for arow in ALTrows:
			newrow = arow[0:3] + ['PAM changed & added'] + arow[4:8] + ["-", "-"] + arow[8:]
			new_guides.append(newrow)
		   # print('impact;pam_changed_added', '-', '->', arow[8], '-', '->', arow[7])

	ALTguides_df = pd.DataFrame(new_guides, columns=columns)
	return ALTguides_df

def create_ALTvcf_report(ALTinfo,altgenome_name):
	columns = ['QueryTerm',f'{altgenome_name} VCF Variant ID',
			   'REF Allele','ALT Examined','ALT allele(s)',
			   'Variant Coordinates','Zygosity','Variant Type','Relative Position to Query']
	rows = []
	for query in ALTinfo.keys():
		#{ALTid : [ALTid,ALTref, ALTalt, ALTalts, ALTcoord, zyg, vtype, rel_pos]}
		for variant, info in ALTinfo[query].items():
			rows.append([query]+info)

	nearby_variants_df = pd.DataFrame(rows,columns=columns)
	return nearby_variants_df


def fetch_ALT_guides(filtered_vcf,
					 guides_report,
					 be_report,
					 guide_search_params,
					 guide_be_search_params,
					 models_path,
					 snv_site_info,
					 diffguides_out,
					 altvar_out,
					 altgenome_name,
					 refgenome_name):

	# Get search parameters and the results from the reference assembly
	# {'spCas9': ('NGG', False, 20, -3, 'requirements work for SpCas9-HF1, eSpCas9 1.1,spyCas9'),
	search_params = pickle.load(open(guide_search_params, 'rb'))
	be_search_params = pickle.load(open(guide_be_search_params, 'rb'))
	if len(be_search_params.values()) == 0:
		be_search_params = None
	# get REF variant info
	# {'X': [['NM_004208.4:c.696+3G>A', 'NM_145812.3', '-', 'AIFM1', '-', 'C', 'T', 'intron', 'GGCTGG...
	hg38_snvinfo = pickle.load(open(snv_site_info, 'rb'))

	print("Searching for ALT genome changes")
	# Chec kto see if there's alt variants in vcf and if so, find guides
	ALTinfo, ALTguides_dict = find_overlapping_variants(filtered_vcf, models_path, hg38_snvinfo, search_params,be_search_params)

	if len(ALTinfo.values()) > 0:

		nearby_variants_df = create_ALTvcf_report(ALTinfo,altgenome_name)
		nearby_variants_df.to_csv(altvar_out, index = False)
		if len(ALTguides_dict.values()) > 0:

			ALTguides_df = guide_compare(guides_report, be_report, ALTguides_dict, ALTinfo, refgenome_name, altgenome_name)
			ALTguides_df.to_csv(diffguides_out,index = False)
		else:

			print(f'no overlapping variants detected in {altgenome_name}')
			with open(diffguides_out, 'w') as f:
				f.write(f"No guide differences found based on the VCF {altgenome_name}")
	else:
		print(f'no overlapping variants detected in {altgenome_name}')
		with open(altvar_out, 'w') as f:
			f.write(f"No Nearby variants found based on the VCF {altgenome_name}")


def main():
	# SNAKEMAKE IMPORTS
	# === Inputs ===
	filtered_vcf = str(snakemake.input.filtered_vcf)
	guides_report = str(snakemake.input.guides_report_out)
	be_report = str(snakemake.input.be_report_out)
	guide_search_params = str(snakemake.input.guide_search_params)
	guide_be_search_params = str(snakemake.input.guide_be_search_params)
	snv_site_info = str(snakemake.input.snv_site_info)
	# === Outputs ===
	diffguides_out = str(snakemake.output.diff_guides)
	altvar_out = str(snakemake.output.alt_var)
	# === Params ===
	idx_filtered_vcf = str(snakemake.params.idx_filtered_vcf)
	# ==* The models_path ideally should not be here.
	#       Once the DataHandler class is properly instantiated this can be removed
	models_path = str(snakemake.params.models_path)
	# === Wildcards ===
	altgenome_name = str(snakemake.wildcards.vcf_id)
	refgenome_name = str(snakemake.wildcards.sequence_id)

	# == Create dummy files for the outputs
	dummy_outputs = [diffguides_out, altvar_out]
	for report in dummy_outputs:
		# Create an empty DataFrame
		df = pd.DataFrame()
		# Export the empty DataFrame to a CSV file
		df.to_csv(str(report))

	# Generate vcf index with tabix
	print(f"Generate tabix file on:\n {idx_filtered_vcf}")
	subprocess.run(f"tabix {filtered_vcf}", shell=True)

	fetch_ALT_guides(filtered_vcf,
					 guides_report,
					 be_report,
					 guide_search_params,
					 guide_be_search_params,
					 models_path,
					 snv_site_info,
					 diffguides_out,
					 altvar_out,
					 altgenome_name,
					 refgenome_name
					 )


if __name__ == "__main__":
	main()



# == Native Modules
import pickle

# == Installed Modules
import pandas as pd


# == Project Modules


def write_out_res(concat_offtarget_scores_dict,
				  concat_guides_report,
				  casoff_params,
				  offtarget_scores_out,
				  formatted_casoff_out):
	'''
	Adds two new columns to guides found table
	Creates and Aggregate summary text output
	'''
	df = pd.DataFrame(concat_offtarget_scores_dict)  # .rename(columns = {'0':'Off Target Score'})
	# this is just a precautionary if this script was run more than once
	if 'Num Off Targets per MM' in concat_guides_report.columns:
		concat_guides_report = concat_guides_report.drop(columns='Num Off Targets per MM')
	if 'Off Target Score' in concat_guides_report.columns:
		concat_guides_report = concat_guides_report.drop(columns='Off Target Score')

	# Update main guides table with specifity Scoress
	spec_scores = df.T.sort_index()['spec_score']
	df = df.drop(index='spec_score')
	df = df.astype('int')
	# Add off_targets summary
	df.index = pd.MultiIndex.from_tuples(list(df.index), names=['BulgeType', 'Number of Mismatches'])
	df = df.reset_index()
	ot_per_mm = df.loc[df['BulgeType'] == 'X']
	ot_per_mm = ot_per_mm.iloc[:, 2:].T.sort_index()
	ot_per_mm = ot_per_mm.stack().astype('str').groupby(level=0).apply('|'.join)
	ot_per_mm = ot_per_mm.rename('Guide_ID')  # ('Num Off Targets per Mismatch')
	concat_guides_report = concat_guides_report.sort_values('Guide_ID')
	concat_guides_report['Off Target Score'] = list(spec_scores)
	concat_guides_report['Num Off Targets per MM'] = list(ot_per_mm)

	# create off_target summary of totals
	if casoff_params[1] == 0:
		df = df.loc[df.BulgeType != 'RNA']
	if casoff_params[2] == 0:
		df = df.loc[df.BulgeType != 'DNA']
	offtarget_summary_report = df.pivot_table(columns=['BulgeType', 'Number of Mismatches'], aggfunc="sum")

	# writeout
	# sum_out = resultsfolder + 'OffTarget_Summary.txt'
	offtarget_summary_report.to_csv(offtarget_scores_out, sep='\t')

	concat_guides_report.to_csv(formatted_casoff_out, index=False)


def main():
	# SNAKEMAKE IMPORTS
	# === Inputs ===
	offtarget_scores_list = list(snakemake.output.offtarget_scores)
	guides_report_per_editor_list = list(snakemake.input.guides_per_editor_path)
	# === Outputs ===
	formatted_casoff_out = str(snakemake.output.formatted_casoff_out)
	offtarget_scores_out = str(snakemake.output.offtarget_scores_out)
	# === Params ===
	rna_bulge = str(snakemake.params.rna_bulge)
	dna_bulge = str(snakemake.params.dna_bulge)
	maximum_mismatches = str(snakemake.params.max_mismatch)
	PU = str(snakemake.params.casoff_accelerator)

	# == Compile Casoff parameters
	casoff_params = (maximum_mismatches, rna_bulge, dna_bulge, PU)

	# == Import per-editor report
	# concatenation using a list of DataFrames
	df_list = []
	for report_path in guides_report_per_editor_list:
		loop_df = pd.read_csv(report_path)
		df_list.append(loop_df)

	# Concatenate all DataFrames at once
	concat_guides_report = pd.concat(df_list)

	# == Import scores dictionary
	concat_offtarget_scores_dict = {}
	for offtarget_scores in offtarget_scores_list:
		with open(offtarget_scores, 'rb') as pickle_handle:
			offtarget_scores_dict = pickle.load(pickle_handle)
			# Update the main dictionary in-place without re-assigning
			concat_offtarget_scores_dict.update(offtarget_scores_dict)

	write_out_res(concat_offtarget_scores_dict,
				  concat_guides_report,
				  casoff_params,
				  offtarget_scores_out,
				  formatted_casoff_out)


if __name__ == "__main__":
	main()
